/*
 * StatDomTree.java
 * - Created by Hiroki Arimura on 10 DEC 01
 * - based on Mattree10.java by Kenji Abe, 2001 for our SDM paper
*/

import java.io.*;
import java.lang.*;
import java.util.*;
import java.math.*;
import java.text.*;
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.openxml.parser.*;
import org.openxml.dom.*;
//import Debug.*;
//import Pattern.*;

public class DomWrapper {

  /************************************************************
   * Global Compilation Switch
   ************************************************************/
  //static final boolean ENABLE_REPORT_TAGTABLE  = true;
  static final boolean ENABLE_REPORT_TAGTABLE  = false;
  //static final boolean ENABLE_REPORT_ATTRTABLE = true;
  static final boolean ENABLE_REPORT_ATTRTABLE = false;
  static final boolean ENABLE_REPORT_TREESTAT  = true;
  //static final boolean ENABLE_REPORT_TREESTAT  = false;
  static final boolean ENABLE_PRINT_TREE       = true;
  //static final boolean ENABLE_PRINT_TREE       = false;

  static final boolean ENABLE_LOGGING_TAGNAMES = false;
  static final boolean ENABLE_LOGGING_ATTRNAMES= false;
  static final boolean ENABLE_DEBUG_PRINT = true;

  /************************************************************
   * Global Compilation Switch
   ************************************************************/
  static Document      domDoc = null;    //Dom document for an input file

  /************************************************************
   * Statistics
   * 
   ************************************************************/
  //Tag Statistics
  static SymbolTable   tagTable  = null; //A tag table

  //Attribute Statistics
  static StringPairMap attrTable = null; //A attribute table
  
  //Tree Statistics 
  static long num_nodes = 0;	
  static long max_depth = 0;	
  static long sum_depth = 0;	
  static long num_leaves = 0;
  static long sum_branch = 0;	
  static long max_branch = 0;

  /************************************************************
   * Main
   * - A test routine for traversing a DOM tree
   ************************************************************/

  public static void main(String args[]){

    // Process arguments
    //System.out.println("##Arg Length = " + args.length);
    if (args.length < 1 || args.length > 1) {
      System.out.println("DomWrapper: too few/many arguments. ");
      System.out.println("Usage: DomWrapper htmlfile");
      System.exit(1);
    }

    File pageFile = new File(args[0]);
    System.out.println("##File name = " + args[0]);
    System.out.println("##File size (byte) = " + pageFile.length());

    try{
      long cputime;
      Debug.task("Initialization...");
      initialize();

      Debug.task("Building an DOM tree by reading the input HTML/XML file...");
      cputime = Debug.checkin();
      Node root    = DomWrapper.createDom(pageFile);
      Document doc = DomWrapper.getDocument();
      Debug.checkout(cputime);

      Debug.task("Report statistics of the original DOM tree...");
      DomWrapper.statTree(root);
      DomWrapper.report();
      Debug.task("END orig");

      Debug.task("Building extended DOM tree...");
      cputime = Debug.checkin();
      DomWrapper.createAttrTree(root, doc);
      Debug.checkout(cputime);

      Debug.task("Report statistics of the original DOM tree...");
      DomWrapper.statTree(root);
      DomWrapper.report();

      if (ENABLE_PRINT_TREE == true) {
	Debug.task("Printing the dom tree...");
	printTree(root, 0);
      }

      tagTable = null; 
      Debug.task("Done.");

    }

    catch(Exception e){
      System.out.println(e);
    }
  }

  /**********************************************************************
   * Dom Wrapper for OpenXML
   * createDom: 
   * - Parse a given input HTML File and then returns the root node
   * - Use instead of YmUtil.parseByOpenXML in YmUtil.java
   **********************************************************************/

  public static Node createDom(File pageFile) throws Exception 
  {
    
    //Start parsing
    FileInputStream in = new FileInputStream(pageFile.toString());
    InputSource input = new InputSource(in);

    //Create a parser
    HTMLParser hp = new HTMLParser();

    //Parse input file
    hp.parse(input);

    //Get DOM object for document
    Document doc = hp.getDocument();
    domDoc = doc; 

    Element root = doc.getDocumentElement();
    return (Node) root; 
  }

  public static Document getDocument()
  {
    return domDoc;
  }

  /**********************************************************************
   * Create extended Dom tree
   * - By attaching attribute-value pairs as element nodes
   * - Attribute nodes appear from left to right in the lexicographic
   * order of their names.
   **********************************************************************/

  /** Main routine for the conversion */
  public static void createAttrTree(Node node, Document doc) 
  {
    String name = node.getNodeName(); 
    appendAttributeNodes(node, doc);

    for (Node next = node.getFirstChild(); next != null;
	 next = next.getNextSibling()){
	  createAttrTree(next, doc);
    }
  }

  /** Utility: 
   *  - insert a new child at the first position */
  static void insertFirstNode(Node node, Node newChild) {
    node.insertBefore(newChild, node.getFirstChild());
  }

  /** Sub-routine for the conversion: 
   * - Sort the attributes of a given node */
  static void appendAttributeNodes(Node node, Document doc)
  {
    String name = node.getNodeName();

    if (ENABLE_LOGGING_ATTRNAMES) { //For debugging
      Debug.println("\t@elem: " + name); 
    }//ENABLE_LOGGING_TAGNAMES

    if (node.getNodeType() == Node.ELEMENT_NODE) {

      //Get a set of attributes for elem
      Element elem = (Element) node; //Downward type casting
      NamedNodeMap map = elem.getAttributes();
      ArrayList attrnameList = new ArrayList();

      //For each attr, do the followings
      for (int i = 0; i < map.getLength(); i++) {
	Attr attr = (Attr) map.item(i);
	String attrname = attr.getNodeName();	//Get attribute name 
	attrnameList.add(attrname);
      }

      //Sort attributes by their names
      Object sortedNames[] = attrnameList.toArray();
      Arrays.sort(sortedNames);

      //Main task: Printing (attr, values) in the sorted order
      for (int i = 0; i < attrnameList.size(); i++) {
	String attrname = (String) sortedNames[i]; 
	String attrvalue = map.getNamedItem(attrname).getNodeValue();
	if (ENABLE_LOGGING_ATTRNAMES) { //For debugging
	  Debug.println("\t\tattr: " + attrname + "\tvalue: " + attrvalue);
	}//ENABLE_LOGGING_TAGNAMES

	//Create a new attribute name node
	Element attrNode = doc.createElement("@" + attrname);
	//Create a new attribute value node
	Element valueNode = doc.createElement("[" + attrvalue + "]");

	//insert at the last position
	//node.appendChild(attrNode); 
	//attrNode.appendChild(valueNode); //insert at the last pos
	//insert at the first pos
	DomWrapper.insertFirstNode(node, attrNode); 
	attrNode.appendChild(valueNode); //insert at the last pos

      }
    }
  } 


  public static void initialize()
  {
    tagTable  = new SymbolTable();   //A tagname table
    attrTable = new StringPairMap(); //An attribute table 

    num_nodes = 0;
    num_leaves= 0;
    max_depth = 0;
    sum_depth = 0;
    max_branch = 0;
    sum_branch = 0;
  }

  public static void report()
  {
    if (ENABLE_REPORT_TAGTABLE == true) {
      Debug.println("Reporting tagTable...");
      tagTable.reportCount();
    }
    if (ENABLE_REPORT_ATTRTABLE == true) {
      Debug.println("Reporting attrTable...");
      attrTable.reportCount();
    }

    if (ENABLE_REPORT_TREESTAT == true) {
      Debug.println("##Stat: Num of nodes  = " + num_nodes);
      Debug.println("##Stat: Num of leaves = " + num_leaves);
      Debug.println("##Stat: Num of tags   = " + tagTable.getSize());
      Debug.println("##Stat: Depth_average  = " +  sum_depth/num_leaves +
		    "\t Depth_maximum = " +  max_depth);
      long num_internals = (num_nodes - num_leaves); 
      Debug.println("##Stat: Branch_average = " +  sum_branch/num_internals +
		    "\t Branch_maximum = " +  max_branch);
    }
  }

  public static void statTree(Node node) 
  {
    tagTable.clear();
    attrTable.clear();
    statTree0(node, 0);
  }

  static void registerAttributes(Node node, StringPairMap attrTable)
  {
    String name = node.getNodeName();
    if (ENABLE_LOGGING_TAGNAMES) { //For debugging
      Debug.println("@node name: " + name +
		    "\tnode type: "+ Node.ELEMENT_NODE);
    }//ENABLE_LOGGING_TAGNAMES

    if (node.getNodeType() == Node.ELEMENT_NODE) {
      Element elem = (Element) node;

      //Get a set of attributes for elem
      NamedNodeMap map = elem.getAttributes();

      //For each attr, do the followings
      for (int i = 0; i < map.getLength(); i++) {
	Attr attr = (Attr) map.item(i);

	//Get attribute name 
	String attrname = attr.getNodeName();

	//Get attribute value
	String attrvalue = attr.getNodeValue();

	//Do something with (attrname, attrvalue)
	int c = attrTable.addPair(attrname, attrvalue);

	//For debugging
	if (ENABLE_LOGGING_TAGNAMES) {
	  int c0 = attrTable.getSounceCount(attrname);
	  Debug.println("\t" + c0 + ":" + c + ":" + 
			"\tattrname=["+ attrname +"]" +
			"\tattrvalue=["+attrvalue+"]");
	}//ENABLE_LOGGING_TAGNAMES

      }
    }
  } 

  private static void statTree0(Node node, int depth) 
  {
    String name = node.getNodeName(); 
    int nid = tagTable.register(name);
    registerAttributes(node, attrTable);
    //appendAttributeNodes(node);

    num_nodes++;

    if  (node.hasChildNodes()){ 
      int branch = 0; 

      for (Node nextnode = node.getFirstChild();
	   nextnode != null;
	   nextnode = nextnode.getNextSibling())
	{
	  statTree0(nextnode, depth + 1);
	  branch = branch + 1;
	}
      max_branch = Math.max(max_branch, branch);
      sum_branch = sum_branch + branch;
    }
    else { 
      num_leaves = num_leaves + 1; 
      max_depth = Math.max(max_depth, depth);
      sum_depth = sum_depth + depth;
    }
  }

  private static void
  printTree(Node node, int depth) 
  {
    String name = node.getNodeName();
    for (int i = 0; i < depth; i++)
      Debug.print("   |");
      Debug.print("----");
    Debug.println(" <" + name + ">");

    if  (node.hasChildNodes()){ 
      for (Node next = node.getFirstChild(); next != null;
	   next = next.getNextSibling()) {
	printTree(next, depth + 1);
      }
    }
    else { 
    }
  }

  public static Node pthParent(Node node, int p)
  { //Note: node is w3c.dom.Node type 
    while (p > 0){
      node = node.getParentNode();
      p--;
    }
    return node;
  }

}

class StringPairMap extends HashMap {
  /** A two level string hash dictionary for pairs of strings (source, dest)
   */
  HashMap sourceMap;
  static String ATTRCOUNT = "@self";

  public StringPairMap()
  {
    super();
    sourceMap = new HashMap();
  }

  private int addDest(HashMap destMap, String dest)
  {
    //Getting destMap
    if (destMap.containsKey(dest) == false) 
      destMap.put(dest, new Integer(0));
    Integer c = (Integer) destMap.get(dest);
    int cnew  = c.intValue() + 1; 
    destMap.put(dest, new Integer(cnew)); //Increment
    return cnew; 
  }

  public int addPair(String source, String dest)
  {
    //Getting sourceMap
    if (this.sourceMap.containsKey(source) == false) 
      this.sourceMap.put(source, new HashMap());
    HashMap destMap = (HashMap) this.sourceMap.get(source);

    //Getting destMap
    int snew = addDest(destMap, StringPairMap.ATTRCOUNT);
    int cnew = addDest(destMap, dest);
    //if (destMap.containsKey(dest) == false) 
    //  destMap.put(dest, new Integer(0));
    //Integer c = (Integer) destMap.get(dest);
    //int cnew  = c.intValue() + 1; 
    //destMap.put(dest, new Integer(cnew)); //Increment
    return cnew; 
  }

  public int getSounceCount(String source)
  { /** Returns whether the edge (source, dest) is contained in
     */
    if (this.sourceMap.containsKey(source)){
      HashMap destMap = (HashMap) this.sourceMap.get(source);
      if (destMap.containsKey(StringPairMap.ATTRCOUNT))
	return ((Integer) destMap.get(StringPairMap.ATTRCOUNT)).intValue();
      else
	return 0;
    }
    else {
      return 0; 
    }
  }

  public int getPairCount(String source, String dest)
  { /** Returns whether the edge (source, dest) is contained in
     */
    if (this.sourceMap.containsKey(source)){
      HashMap destMap = (HashMap) this.sourceMap.get(source);
      if (destMap.containsKey(dest))
	return ((Integer) destMap.get(dest)).intValue();
      else
	return 0;
    }
    else {
      return 0; 
    }
  }

  public HashMap getSourceMap()
  {
    return (HashMap) this.sourceMap; 
  }

  public HashMap getDestMap(String source)
  {
    if (this.sourceMap.containsKey(source)){
      return (HashMap) this.sourceMap.get(source);
    }
    else
      return null;
  }

  public void reportCount()
  {
    HashMap sourceMap = (HashMap) this.sourceMap; 

    //Process sourceMap
    Set entries = sourceMap.entrySet();
    Iterator iter = entries.iterator();
    while (iter.hasNext()) {
      Map.Entry entry = (Map.Entry) iter.next();
      String source    = (String) entry.getKey();
      //Do something
      Debug.println("\tsource: " + source); 

      HashMap destMap = (HashMap) entry.getValue();

      //Process destMap
      Set d_entries = destMap.entrySet();
      Iterator d_iter = d_entries.iterator();
      while (d_iter.hasNext()) {
	Map.Entry d_entry = (Map.Entry) d_iter.next();
	String dest  = (String) d_entry.getKey();
	int count = ((Integer) d_entry.getValue()).intValue();
	//Do something
	Debug.println("\t" + count + "\tdest: " + dest);
      } // while dest
    } //while source
  }

}//StringPairMap
